#!/usr/bin/sbcl
(eval-when (:compile-toplevel :load-toplevel :execute)
  (require 'asdf)
  (setf *invoke-debugger-hook* (lambda (condition hook)
				 (declare (ignore hook))
				 #+sbcl(sb-debug:backtrace 20 *error-output*)
				 #+cmu(debug:backtrace 20 *error-output*)
				 (format *error-output* "Error: ~A~%" condition)
				 (quit)))
  ; load current directory's ASD file first
  (push '*default-pathname-defaults* asdf:*central-registry*)
  (asdf:oos 'asdf:load-op :araneida))

#|
This script is the server side of the basic Araneida test suite

important bookmarks:
defvar *url*
defvar *listener*
defun get-random-text
:teststart
:testend
start-listening

urls have a bookmark:

url: <absolute url minus host>

for example: "url: /strict-match/"

|#

(defpackage "TEST-LISTENER"
  (:use "COMMON-LISP" "ARANEIDA"))

(in-package test-listener)
(defvar *quit* 0)

(defvar *url* (parse-urlstring "http://localhost:15840"))
(defvar *listener* (make-instance #+araneida-threads 'threaded-http-listener
				  #-araneida-threads 'serve-event-http-listener
				  :address #(0 0 0 0)
				  :default-hostname "localhost"
				  :port (url-port *url*)))

(defun get-random-text ()
  "Returns the random text passed at the command line.
If it can't be figured out, it defaults to '42'"
  (or #+sbcl(second sb-ext:*posix-argv*)
      #+cmu(fifth extensions:*command-line-strings*)
      "42"))

(defvar *uniquetext* (format nil "|RANDOMID=~A|" (get-random-text))
  "This unique identifier should be inserted into every page. This is so we know we're
not connecting to a stale session.")

(defun gencontent (string)
  (declare (type (or string cons) string))
  `(html (body
	  (h1 ,string)
	  (h1 ,*uniquetext*))))

(defmacro simplegen (string)
  (let ((string-val (gensym)))
    `(let ((,string-val ,string))
      (request-send-headers request)
      (html-stream (request-stream request)
       (gencontent ,string-val)))))

(defmacro my-install-handler (class url &optional (strict t))
  `(install-handler (http-listener-handler *listener*)
    (make-instance ',class)
    (urlstring (merge-url *url* ,url)) ,strict))
(defmacro install-handler-for (obj url &optional (strict t))
  `(install-handler (http-listener-handler *listener*)
    ,obj
    (urlstring (merge-url *url* ,url)) ,strict))

#| :teststart - the start of tests |# 

#|  -------- url: /handler   ------------------  |#
(defclass simple-handler (araneida:handler) ())
(defmethod handle-request-response ((handler simple-handler) method request)
  (simplegen "simple handler"))
(my-install-handler simple-handler "/handler")

#|  -------- url: /strict-match/   ------------------  |#
(defclass strict-matcher (araneida:handler) ())
(defmethod handle-request-response ((handler strict-matcher) method request)
  (simplegen "strict matcher"))
(my-install-handler strict-matcher "/strict-match/")

#|  -------- url: /loose-match/   ------------------  |#
(defclass loose-matcher (araneida:handler) ())
(defmethod handle-request-response ((handler loose-matcher) method request)
  (simplegen "loose matcher"))
(my-install-handler loose-matcher "/loose-match/" nil)

#|  -------- url: /dispatch-handler/   ------------------  |#
; test a dispatching handler - and test to make sure it can handle its own response
(defclass dispatch-handler (araneida:dispatching-handler) ())
(defmethod handle-request-response :around ((handler dispatch-handler) method request)
  (or (call-next-method)
      (simplegen "dispatch handler")))
(defclass dispatch-handler-subhandler (araneida:handler) ())
(defmethod handle-request-response ((handler dispatch-handler-subhandler) method request)
  (simplegen "dispatch subhandler handler"))
(defmethod shared-initialize :after ((handler dispatch-handler) slot-names &rest initargs)
  (declare (ignore slot-names initargs))
  (install-handler handler (make-instance 'dispatch-handler-subhandler) "subhandler" t))

(install-handler (http-listener-handler *listener*) (make-instance 'dispatch-handler)
		 (urlstring (merge-url *url* "/dispatch-handler/")) nil)

#|  -------- url: /error/   ------------------  |#
(defclass generate-error (araneida:handler)
  ((code :initarg :code :reader generate-error-code)))
(defmethod handle-request-response ((handler generate-error) method request)
  (error (generate-error-code handler) :client-message (format nil "Test error. ~A" *uniquetext*)))
(macrolet ((install-error-handler (error-name)
	     (let* ((symname (symbol-name error-name))
		    (error-sym (intern (format nil "HTTP-~A" symname) (find-package "ARANEIDA")))
		    (error-url (format nil "/error/~A" (string-downcase symname))))
	       `(install-handler-for (make-instance 'generate-error :code ',error-sym) ,error-url))))
  (install-error-handler bad-request)
  (install-error-handler unauthorized)
  (install-error-handler payment-required)
  (install-error-handler forbidden)
  (install-error-handler payment-required)
  (install-error-handler not-found)
  (install-error-handler method-not-allowed)
  (install-error-handler not-acceptable)
  (install-error-handler proxy-authentication-required)
  (install-error-handler request-time-out)
  (install-error-handler conflict)
  (install-error-handler gone)
  (install-error-handler length-required)
  (install-error-handler precondition-failed)
  (install-error-handler request-entity-too-large)
  (install-error-handler request-url-too-large)
  (install-error-handler unsupported-media-type)
  (install-error-handler internal-server-error)
  (install-error-handler not-implemented)
  (install-error-handler bad-gateway)
  (install-error-handler service-unavailable)
  (install-error-handler gateway-time-out)
  (install-error-handler version-not-supported))

#|  -------- url: /redirect/first   ------------------  |#
#|  -------- url: /redirect/second   ------------------  |#
(defclass redirect-first-handler (araneida:handler) ())
(defmethod handle-request-response ((handler redirect-first-handler) method request)
  (request-redirect request (merge-url *url* "/redirect/second")))
(my-install-handler redirect-first-handler "/redirect/first")

(defclass redirect-second-handler (araneida:handler) ())
(defmethod handle-request-response ((handler redirect-second-handler) method request)
  (simplegen "redirect second handler"))
(my-install-handler redirect-second-handler "/redirect/second")

#|  -------- url: /redirect-handler/first   ------------------  |#
#|  -------- url: /redirect-handler/second   ------------------  |#
; we don't have to subclass here for the redirect handler, it just helps with error checking
(defclass redirect-handler-first-handler (araneida:redirect-handler)
  ((araneida::location :initform (merge-url *url* "/redirect-handler/second"))))
(my-install-handler redirect-handler-first-handler "/redirect-handler/first")
(defclass redirect-handler-second-handler (araneida:handler) ())
(defmethod handle-request-response ((handler redirect-handler-second-handler) method request)
  (simplegen "redirect-handler second handler"))
(my-install-handler redirect-handler-second-handler "/redirect-handler/second")

#|  -------- url: /cookie/send   ------------------  |#
#|  -------- url: /cookie/receive   ------------------  |#
; SCORE! WWW::Mechanize cannot handle paths
(defclass cookie-send (araneida:handler) ())
(defmethod handle-request-response ((handler cookie-send) method request)
  (let ((cookies
	 (list
	  (cookie-string "simplecookie" "simple-value")
	  (cookie-string "ladencookie" "123ladenvalue123---" :comment "a comment!" :max-age 30))))
    (request-redirect request (merge-url *url* "/cookie/receive") :set-cookie cookies)))
(my-install-handler cookie-send "/cookie/send")
(defclass cookie-receive (araneida:handler) ())
(defmethod handle-request-response ((handler cookie-receive) method request)
  (let ((cookies '("simplecookie" "ladencookie")))
    (simplegen (format nil "cookie receive handler.<br>~%~:{unsafecookie ~A=~A<br>~%~}~%~:{safecookie ~A=~A<br>~%~}~%"
		       (mapcar (lambda (cookie-name) (list cookie-name (request-cookie request cookie-name)))
			       cookies)
		       (mapcar (lambda (cookie-name) (list cookie-name (request-safe-cookie request cookie-name nil)))
			       cookies)))))
(my-install-handler cookie-receive "/cookie/receive")


#|  -------- url: /conditional-get   ------------------  |#
(defclass conditional-get-handler (araneida:handler) ())
(defmethod handle-request-response ((handler conditional-get-handler) method request)
  (request-send-headers request
			:conditional t
			:last-modified (encode-universal-time 0 0 0 1 1 1990 0))
  (html-stream (request-stream request)
	       (gencontent "conditional get handler")))
(my-install-handler conditional-get-handler "/conditional-get")


#|  -------- url: /paramtest   ------------------  |#
(defclass paramtest-handler (araneida:handler) ())
(defmethod handle-request-response ((handler paramtest-handler) method request)
  (let ((params (url-query-alist (request-url request)))
	(foo (url-query-param (request-url request) "foo"))
	(foo-nocase (url-query-param (request-url request) "FOO" :case-sensitive nil))
	(foo-taint (tainted-url-query-param (request-url request) "foo"))
	(foo-taint-nocase (tainted-url-query-param (request-url request) "FOO" :case-sensitive nil)))
    (simplegen `(div
		 (p "foo-alist " ,(format nil (second (assoc "foo" params :test #'string-equal))))
		 (p "foo untaint case " ,(car foo))
		 (p "foo untaint nocase " ,(car foo-nocase))
		 (p "foo taint case " ,(untaint (lambda (x) x) (car foo-taint))) ; normally, this would check to make sure all the elements are something
		 (p "foo taint nocase " ,(untaint (lambda (x) x) (car foo-taint-nocase)))
		 (p "foo type taint case " ,(type-of (car foo-taint)))
		 (p "foo type taint nocase " ,(type-of (car foo-taint-nocase)))))))
(my-install-handler paramtest-handler "/paramtest" nil)


#|  -------- url: /mime/text-plain   ------------------  |#
(defclass text-plain-handler (araneida:handler) ())
(defmethod handle-request-response ((handler text-plain-handler) method request)
  (send-file request "NEWS" :content-type "text/plain"))
(my-install-handler text-plain-handler "/mime/text-plain")


#|  -------- url: /authenticated/test   ------------------  |#
(defclass authenticated-test-handler (araneida:handler) ())
(defun authme (username password)
  (if (and (equal username "testuser")
	   (equal password "testpass"))
      1
      nil))
(defmethod handle-request-authentication ((handler authenticated-test-handler) method request)
  (funcall (araneida::make-basic-authentication-handler "testrealm" #'authme)
	   request nil))
(defmethod handle-request-response ((handler authenticated-test-handler) method request)
  (simplegen "authenticated test handler"))
(my-install-handler authenticated-test-handler "/authenticated/test")


#|  -------- url: /authenticated/test   ------------------  |#
(defclass authorization-test-handler (araneida:handler) ())
(defmethod handle-request-authorization ((handler authorization-test-handler) method request)
  (if (equal (first (request-header request :referer))
	     "http://example.com/foobar")
      t
      (signal 'http-unauthorized)))
(defmethod handle-request-response ((handler authorization-test-handler) method request)
  (simplegen "authorization test handler"))
(my-install-handler authorization-test-handler "/authorization/test")


#|  -------- url: /quit   ------------------  |#
(defclass quit-handler (araneida:handler) ())
(defmethod handle-request-response ((handler quit-handler) method request)
  (simplegen "quitting")
  (setf *quit* 1)
  (stop-listening *listener*))
(my-install-handler quit-handler "/quit")


#| :testend - the end of test servers |#
(start-listening *listener*)
(loop while (eql *quit* 0)
      do #+araneida-threads(sleep 1)
         #-araneida-threads(araneida::serve-event-http-listener-accept-one-request *listener*))
#+sbcl(sb-ext:quit)
#+cmu(extensions:quit)
